#include "control/specifications.h"

/**
 *Gets position and velocity info, 
 * Sending thrust info to Morse (and Vehicle(ROS) for verification of MORSE model)
 */

Matrix<float, 7, 1> eta_q;
Matrix<float, 6, 1> eta, eta_desired;

Matrix<float, Dynamic, 1> controlPerThruster;
Matrix<float, 6, Dynamic> T;
int nThrusters = 0;

Matrix<float, 6, 1> nu;



//Called when position info is updated
void updatePosition(const geometry_msgs::PoseStamped::ConstPtr& msg)
{
  eta_q(0) = msg->pose.position.x;
  eta_q(1) = msg->pose.position.y;
  eta_q(2) = msg->pose.position.z;

  eta_q(3) = msg->pose.orientation.w;
  eta_q(4) = msg->pose.orientation.x;
  eta_q(5) = msg->pose.orientation.y;
  eta_q(6) = msg->pose.orientation.z;

  //Finding euler angles from quaternions
  eta=updateEta(eta_q);

}

void updateVelocity(const geometry_msgs::TwistStamped::ConstPtr& msg) 
{
  nu(0) = msg->twist.linear.x;
  nu(1) = msg->twist.linear.y;
  nu(2) = msg->twist.linear.z;

  nu(3) = msg->twist.angular.x;
  nu(4) = msg->twist.angular.y;
  nu(5) = msg->twist.angular.z;
//cout << "Got Velocity:   " << nu.transpose() << endl;
}


//updates thrust message when new thrust is calculated
void updateRPMmsg(std_msgs::Float32MultiArray* RPM_msg)
{
  //Vector length depends on number of thrusters
  //Adds new element to vector for each thruster
  for(int i=0; i<nThrusters; i++)
    {
      RPM_msg->data.push_back(controlPerThruster(i));
    }
}


int main(int argc, char **argv)
{
  cout << "############ DP Controller ############\n" << endl;  

  ros::init(argc, argv, "DPcontroller");

  ros::NodeHandle n;


  //Get thrust allocation matrix
  ifstream specifications;
  T=getThrustAllocation(specifications, &nThrusters);
  controlPerThruster.resize(nThrusters, 1);
  T.resize(6, nThrusters);

  //Get path of dynamics
  string posePathString, twistPathString;
  posePathString=getPoseTopic(specifications);
  twistPathString=getTwistTopic(specifications);
  specifications.close();

  //To publish thrust info
  std_msgs::Float32MultiArray RPM_msg;
  ros::Publisher controller_pub = n.advertise<std_msgs::Float32MultiArray>("RPMcontrol/DP", 1);

  //Subscribe for vehicle positions
  ros::Subscriber sub_ROVpos = n.subscribe(posePathString,1, updatePosition);

  //Subscribe for vehicle velocity
  ros::Subscriber sub_ROVvel = n.subscribe(twistPathString,1, updateVelocity);


  ros::Rate loop_rate(N);

  Matrix<float, 6, 1> error, proportional_control, derivative_control, integral_control, error_integral, control_param, control0;
  error_integral << 0.0, 0.0, 0.0, 0.0, 0.0, 0.0;

  //POSITION FOR DP
  eta_desired << 0.0, 0.0, 5.0, 0.0, 0.0, 0.0;
  eta=eta_desired;  //To not produce any thrust before true values of eta are subscribed to

  //Control gain matrices
  Matrix<float, 6, 6> K_p, K_d, K_i;

  K_p <<  1000, 0, 0, 0, 0, 0,
          0, 1000, 0, 0, 0, 0, 
          0, 0, 1000, 0, 0, 0, 
          0, 0, 0, 0, 0, 0, 
          0, 0, 0, 0, 0, 0, 
          0, 0, 0, 0, 0, 200;

  K_d <<  10, 0, 0, 0, 0, 0,
          0, 10, 0, 0, 0, 0, 
          0, 0, 10, 0, 0, 0, 
          0, 0, 0, 0, 0, 0, 
          0, 0, 0, 0, 0, 0, 
          0, 0, 0, 0, 0, 500;

  K_i <<  50, 0, 0, 0, 0, 0,
          0, 50, 0, 0, 0, 0, 
          0, 0, 50, 0, 0, 0, 
          0, 0, 0, 0, 0, 0, 
          0, 0, 0, 0, 0, 0, 
          0, 0, 0, 0, 0, 10; 

  //To make up for boyancy force from beginning (not bother integral part with this, as it is known information)
  //control0 << 0.0, 0.0, 328.0, 0.0, 0.0, 0.0;

  float RPMmax=1450;
  float RPMmin=-1450;

  Vector3f euler;
  euler << 0, 0, 0;
  Matrix3f R_mat;
  Matrix<float, 6, 6> R;
  R <<  1, 0, 0, 0, 0, 0,
        0, 1, 0, 0, 0, 0,
        0, 0, 1, 0, 0, 0,
        0, 0, 0, 1, 0, 0, 
        0, 0, 0, 0, 1, 0,
        0, 0, 0, 0, 0, 1;


  //Keep heading error between -PI:PI
  while (ros::ok())
  {
    //Find pose error 
    error= eta-eta_desired;

  if (error(5)>PI)
  {
    error(5)=error(5)-2*PI;
  }
  else if (error(5)<-PI)
  {
    error(5)=error(5)+2*PI;
  }


    //Update integrals
    error_integral=error_integral+error*dt;

    euler(2)=eta(5);
    R_mat=rotationMatrix(euler);
    
  //TODO fix (implement with less code :) )
    R(0, 0) = R_mat(0, 0);
    R(1, 0) = R_mat(1, 0);
    R(2, 0) = R_mat(2, 0);
    R(0, 1) = R_mat(0, 1);
    R(1, 1) = R_mat(1, 1);
    R(2, 1) = R_mat(2, 1);
    R(0, 2) = R_mat(0, 2);
    R(1, 2) = R_mat(1, 2);
    R(2, 2) = R_mat(2, 2);
    

    //Find control parameters
    proportional_control=-R.transpose()*K_p*error;
    derivative_control=K_d*R*nu;
    integral_control=-R.transpose()*K_i*error_integral;

    control_param=proportional_control+derivative_control+integral_control;

    //Find RPM per thruster
    controlPerThruster=T.colPivHouseholderQr().solve(control_param);

    //Saturation
    for (int i=0; i<nThrusters; i++)
    {
        controlPerThruster(i)=min(max(controlPerThruster(i), RPMmin), RPMmax);
    }

  
    //Publish thrust
    RPM_msg.data.clear();
    updateRPMmsg(&RPM_msg);
    controller_pub.publish(RPM_msg);


    ros::spinOnce();
    loop_rate.sleep();
  }
  return 0;
}

