#include "control/specifications.h"

/**
 *Gets guidance info, 
 * Sending thrust info to Morse (and Vehicle(ROS) for verification of MORSE model)
 */

//Matrices and div
Matrix<float, 7, 1> eta_q;
Matrix<float, 6, 1> eta;

Matrix<float, 6, 1> control_param;
Matrix<float, Dynamic, 1> controlPerThruster;
Matrix<float, 6, Dynamic> T;
int nThrusters = 0;

float heading_desired;
Matrix<float, 6, 1> nu;
float depth_desired;
float speed_desired=0;

//Thrust limits
float RPMmax=1450;
float RPMmin=-1450;

float int_limit=100;
float heading_integral, depth_integral, speed_integral;
float heading_error, speed_error, depth_error;


//Called when position info is updated
void updatePosition(const geometry_msgs::PoseStamped::ConstPtr& msg)
{
  eta_q(0) = msg->pose.position.x;
  eta_q(1) = msg->pose.position.y;
  eta_q(2) = msg->pose.position.z;

  eta_q(3) = msg->pose.orientation.w;
  eta_q(4) = msg->pose.orientation.x;
  eta_q(5) = msg->pose.orientation.y;
  eta_q(6) = msg->pose.orientation.z;

  //Finding euler angles from quaternions
  eta=updateEta(eta_q);

}

void updateVelocity(const geometry_msgs::TwistStamped::ConstPtr& msg)  //TODO Use in controller
{
  nu(0) = msg->twist.linear.x;
  nu(1) = msg->twist.linear.y;
  nu(2) = msg->twist.linear.z;

  nu(3) = msg->twist.angular.x;
  nu(4) = msg->twist.angular.y;
  nu(5) = msg->twist.angular.z;
}


//updates thrust message when new thrust is calculated
void updateRPMmsg(std_msgs::Float32MultiArray* RPM_msg)
{
  //Vector length depends on number of thrusters
  //Adds new element to vector for each thruster
  for(int i=0; i<nThrusters; i++)
    {
      RPM_msg->data.push_back(controlPerThruster(i));
    }
}


//Called when guidance publishes new info
void updateControl(const guidance::headingGuidance::ConstPtr& msg)
{
  heading_desired=msg->desiredHeading;
  depth_desired=msg->desiredDepth;

  //Keep heading error between -PI:PI
  heading_error=eta(5)-heading_desired;
  if (heading_error>PI)
  {
    heading_error=heading_error-2*PI;
  }
  else if (heading_error<-PI)
  {
    heading_error=heading_error+2*PI;
  }

  depth_error=eta(2)-depth_desired;
  float speed_tot=sqrt(nu(0)*nu(0)+nu(1)*nu(1));

  //Calculate control parameters in each direction
  //Assume open stability in roll and pitch; heading controller controls yaw force, depth controller controls z-force, speed controller controls x-force. y, p and q set to zero. 
  control_param(0)=speedController(&speed_desired, speed_tot, speed_integral, heading_error);
  control_param(1)=0;
  control_param(2)=depthController(depth_desired, eta(2), depth_integral);
  control_param(3)=0;
  control_param(4)=0;
  control_param(5)=headingController(heading_desired, eta(5), heading_integral, nu(5));
 
  speed_error=nu(0)-speed_desired;

  //Find desired RPM for each thruster
  controlPerThruster=T.colPivHouseholderQr().solve(control_param);
  

  //For troubleshooting. Overwriting controller, setting each RPM directly  
  controlPerThruster(0)=0;
  controlPerThruster(1)=0;
  controlPerThruster(2)=0;
  controlPerThruster(3)=1000;
  controlPerThruster(4)=700;


  bool scaleYaw = false;
  bool scaleSpeed=false;
  bool tir=true;


// Ad Hoc method to saturate RPM values without loosing yaw moment in the process. TODO Find better method
  while (tir)
  {
    if (controlPerThruster(0)>RPMmax || controlPerThruster(0) <RPMmin)
    {
      scaleYaw=true;
    }
    else if (controlPerThruster(3)>RPMmax || controlPerThruster(4)>RPMmax || controlPerThruster(3)<RPMmin || controlPerThruster(4)<RPMmin)
    {
      scaleSpeed=true;
      scaleYaw=true;
    }
    else if (controlPerThruster(1)>RPMmax || controlPerThruster(1) <RPMmin)
    {
      controlPerThruster(1)=min(max(controlPerThruster(1), RPMmin), RPMmax);
      controlPerThruster(2)=min(max(controlPerThruster(2), RPMmin), RPMmax);
    }
    else
    {
      tir=false;
    }

    if (scaleSpeed==true && scaleYaw==true)
    {
      control_param(0)=0.9*control_param(0);
      control_param(5)=0.9*control_param(5);
    }
    if (scaleYaw==true && scaleSpeed==false)
    {
      control_param(5)=0.9*control_param(5);
    }

    if (scaleYaw==true || scaleSpeed==true)
    {
      controlPerThruster=T.colPivHouseholderQr().solve(control_param);
      scaleYaw=false;
      scaleSpeed=false;
    }
  }
  return;
}


int main(int argc, char **argv)
{
  cout << "############ Path Controller ############\n" << endl;  

  ros::init(argc, argv, "pathController");

  ros::NodeHandle n;


  //Get thrust allocation matrix
  ifstream specifications;
  T=getThrustAllocation(specifications, &nThrusters);
  controlPerThruster.resize(nThrusters, 1);
  T.resize(6, nThrusters);
  for(int i; i<nThrusters; i++)
  {
    controlPerThruster(i)=0;
  }

  //Get path of dynamics
  string posePathString, twistPathString;
  posePathString=getPoseTopic(specifications);
  twistPathString=getTwistTopic(specifications);

  //Get path for guidance
  string guidancePathString;
  guidancePathString=getGuidanceTopic(specifications);
  specifications.close();

  //To publish control info
  std_msgs::Float32MultiArray RPM_msg;
  ros::Publisher controller_pub = n.advertise<std_msgs::Float32MultiArray>("RPMcontrol/Path", 1);

  //Subscribe for vehicle positions
  ros::Subscriber sub_ROVpos = n.subscribe(posePathString,1, updatePosition);

  //Subscribe for vehicle velocity TODO morse
  ros::Subscriber sub_ROVvel = n.subscribe(twistPathString,1, updateVelocity);

  //Subscribe for guidance
  ros::Subscriber sub_guidance = n.subscribe(guidancePathString,1, updateControl);

  ros::Rate loop_rate(N);

  depth_integral=0.0;
  speed_integral=0.0;
  heading_integral=0.0;


  while (ros::ok())
  {
   //Update integrals TODO saturation
    depth_integral=depth_integral+depth_error*dt;
    speed_integral=speed_integral+speed_error*dt;
    heading_integral=heading_integral+heading_error*dt;

    if (speed_integral>int_limit || speed_integral <-int_limit)
    {
      speed_integral=min(max(speed_integral, -int_limit), int_limit);
    }


    //Publish thrust
    RPM_msg.data.clear();
    updateRPMmsg(&RPM_msg);
    controller_pub.publish(RPM_msg);

    ros::spinOnce();
    loop_rate.sleep();
  }
  return 0;
}

