#include "dynamics/specifications.h"


//Run everything with <bash executeShell> in terminal

/**
 Main file for ROV dynamics. 
 Takes damping(linear+quadratic), coriolis, added mass and restoring forces, together with a thrust forces, and finds changes in motion. 

TODO   
	Change structure; object oriented?	 
  Delete variables not used (anymore)?
  Better implementation for current speed/direction/profile
  wave forces
**/


  Matrix<float, Dynamic, 1> thrustPerThruster, thrusterRPM, thruster_diameter;
  Matrix<float, 6, 1> thrustBody, nu_rel;
  Matrix<float, 6, Dynamic> T;
  Matrix<float, 2, 4> thrust_coeff;
  Matrix<float, 2, Dynamic> thrustloss_factors, RPM_limits;
  float nThrusters = 0;


//Called when controller publishes info
void updateThrust(const std_msgs::Float32MultiArray::ConstPtr& msg)
{
  //Get data, RPM for each thruster
  for(int i; i<nThrusters; i++)
  {
    thrusterRPM(i) = msg->data[i];
  }

  //Estimates thrust in each thruster from desired RPM
  thrustPerThruster=thrustFromRPM(thrusterRPM, nThrusters, thrust_coeff, nu_rel, thruster_diameter, thrustloss_factors, RPM_limits);

  //Calculate total force and torque in body plane based on thrust in each thruster
  thrustBody=T*thrustPerThruster;
  //thrustBody << 100, 0, 0, 0, 0, 0; //For overwriting controller (troubleshooting)
}

//For publishing calculated ROV pose
void updatePosmsg(geometry_msgs::PoseStamped* msg, Matrix<float, 7, 1>& eta_q)
{
  msg->pose.position.x=eta_q(0); 
  msg->pose.position.y=eta_q(1); 
  msg->pose.position.z=eta_q(2);

  msg->pose.orientation.w=eta_q(3);
  msg->pose.orientation.x=eta_q(4);
  msg->pose.orientation.y=eta_q(5);
  msg->pose.orientation.z=eta_q(6);
}


//For publishing calculated ROV velocity
void updateVelmsg(geometry_msgs::TwistStamped* msg, Matrix<float, 6, 1>& nu)
{
  msg->twist.linear.x=nu(0); 
  msg->twist.linear.y=nu(1); 
  msg->twist.linear.z=nu(2);

  msg->twist.angular.x=nu(3);
  msg->twist.angular.y=nu(4);
  msg->twist.angular.z=nu(5);
}


int main(int argc, char **argv)
{ 
  cout << "#########  Vehicle Dynamics ROS ########## \n" <<  endl;

  ros::init(argc, argv, "Vehicle");

  ros::NodeHandle n;

  ros::Rate loop_rate(N);

  //Current info. For now only uniform current. 
  //TODO: Better implementation (take current as input somehow) 
  //      Not only uniform, include also other profiles  
  float current_direction=PI;
  float current_speed=0.0;

  //Vector for storing current velocity relative to body coordinate system
  Matrix<float, 6, 1> current_bodycoor;
  current_bodycoor << 0.0, 0.0, 0.0, 0.0, 0.0, 0.0;


  //Div vectors 
  Matrix<float, 6, 1> damping_force, eta, G_restoring, nu;
  Matrix<float, 7, 1> eta_q, eta_q_dot;
  Vector3f r_g, r_b;
  nu_rel << 0, 0, 0, 0, 0, 0;

  //Matrices
  Matrix<float, 6, 6> M_tot, M_inv, M_rb, M_a, D_l, D_q;
  Matrix<float, 6, 12> Coriolis;
  Matrix<float, 13, 1> motion;

  //Constants
  float V, W, B;


  //Get info about the ROV, reads from inputFile.txt
  ifstream specifications;
  openInputFile(specifications);

  eta=getInitialEta(specifications);
  nu=getInitialNu(specifications);
  M_rb=getMass(specifications);
  M_a=getAddedMass(specifications);
  D_l=getLinearDampingMatrix(specifications);
  D_q=getQuadraticDampingMatrix(specifications);
  r_g=getCOG(specifications);
  r_b=getCOB(specifications);
  V=getVolume(specifications);
  T=getThrustAllocationMatrix(specifications, &nThrusters);
  thruster_diameter=getThrusterDiameter(specifications, nThrusters);
  thrust_coeff=getThrustCoeffs(specifications, nThrusters);
  thrustloss_factors=getThrusterlossFactors(specifications, nThrusters);
  RPM_limits=getLimitsRPM(specifications, nThrusters);



  //Resize vectors now that number of thrusters is known
  thrustPerThruster.resize(nThrusters, 1);
  thrusterRPM.resize(nThrusters, 1);
  T.resize(6, nThrusters);
  thruster_diameter.resize(nThrusters, 1);
  thrustloss_factors.resize(2, nThrusters);
  RPM_limits.resize(2, nThrusters);
  

  //Read topic info from inputFile.txt
  string controllerTopic;
  controllerTopic=getControllerTopic(specifications);
  specifications.close();


  //Get pose in quaternions
  eta_q=updateEtaQ(eta);

  //Find total buoyancy force and gravity force
  W=g*M_rb(0, 0);
  B=g*rho*V;
  

  // Find M_tot = M_a + M_rb and inverse of M_tot
  M_tot=M_a+M_rb;
  M_inv=M_tot.inverse();

  //Subscribe for control info
  ros::Subscriber sub_control = n.subscribe(controllerTopic, 1, updateThrust);

  //Publish to whoever wants to know position and orientation of the ROV
  geometry_msgs::PoseStamped Pos_msg;
  ros::Publisher pub_Pos = n.advertise<geometry_msgs::PoseStamped>("Dynamics/ROS/Pose",1000);

  //Velocity
  geometry_msgs::TwistStamped Vel_msg;
  ros::Publisher pub_vel = n.advertise<geometry_msgs::TwistStamped>("Dynamics/ROS/Twist",1000);
 

 //for feilsøking i krefter
  double time_relative, time_init;

  ofstream lin_damping;
  ofstream quad_damping;
  ofstream coriolis_a;
  ofstream coriolis_rb;
  ofstream restoring;
  ofstream thrust_body;
  lin_damping.open("lin_damping_ros.txt");
  quad_damping.open("quad_damping_ros.txt");
  coriolis_a.open("coriolis_a_ros.txt");
  coriolis_rb.open("coriolis_rb_ros.txt");
  restoring.open("restoring_ros.txt");
  thrust_body.open("thrust_body.txt"); 

  ros::Duration(3.36).sleep();
  time_init=ros::Time::now().toSec(); 



  while (ros::ok())
  {

    time_relative=ros::Time::now().toSec()-time_init;


    //Find current velocity relative to body coordinate system
    current_bodycoor(0, 0)=current_speed*cos(current_direction-eta(5, 0));
    current_bodycoor(1, 0)=current_speed*sin(current_direction-eta(5, 0));
    
   
    //Find relative velocity of the ROV
    nu_rel=nu-current_bodycoor;

    //Find Coriolis Matrices (Coriolis=[C_rb, C_a])
	  Coriolis=updateCoriolis(nu, nu_rel, M_rb, M_a, r_g);
    
    //Find restoring forces
 	  G_restoring=updateRestoring(W, B, eta, r_g, r_b);
    restoring << time_relative<< "\t" << -1*G_restoring.transpose() << "\n";
    thrust_body << time_relative<< "\t" << thrustBody.transpose() << "\n";


    //Find motion of ROV from total forces and thrust
	  motion=updateMotion(M_inv, D_l, D_q, Coriolis, G_restoring, nu, nu_rel, thrustBody, eta_q, lin_damping, quad_damping, coriolis_a, coriolis_rb, time_relative); //TODO: If we don't want the calculated forces to be stored, the input values for the updateMotion function should be removed
    eta_q = motion.topLeftCorner(7, 1);
    nu=motion.bottomLeftCorner(6, 1);

    //Find eta from eta_q
    eta=updateEta(eta_q);

    //Publish info about position and orientation of ROV
    updatePosmsg(&Pos_msg, eta_q);
    pub_Pos.publish(Pos_msg);

    //Publish info about velocity of ROV
    updateVelmsg(&Vel_msg, nu);
    pub_vel.publish(Vel_msg);


    ros::spinOnce();
    loop_rate.sleep();
  }

  //Closing files
  lin_damping.close();
  quad_damping.close();
  coriolis_a.close();
  coriolis_rb.close();
  restoring.close();
  thrust_body.close();

  return 0;
}

